import unittest

from .. import base


class TestContext(unittest.TestCase):

    def test_basic_usage(self):
        ctx = base.Context()
        ctx['something'] = 'test'

        self.assertEqual(ctx['something'], 'test')

    def test_global_exists(self):
        ctx = base.Context()

        self.assertEqual(ctx['global'], {})

    def test_context_manager(self):
        ctx = base.Context()
        ctx['something'] = 'test'
        ctx['over'] = 'test'

        with ctx as ictx:
            ictx['else'] = 'else'
            ictx['over'] = 'over'

        self.assertNotIn('else', ctx)
        self.assertEqual(ctx['something'], 'test')
        self.assertEqual(ictx['over'], 'over')
        self.assertEqual(ctx['over'], 'test')

    def test_context_manager_copywith(self):
        ctx = base.Context()
        ctx['something'] = 'test'
        ctx['over'] = 'test'

        with ctx.copywith() as ictx:
            ictx['else'] = 'else'
            ictx['over'] = 'over'

        self.assertNotIn('else', ctx)
        self.assertEqual(ctx['something'], 'test')
        self.assertEqual(ictx['over'], 'over')
        self.assertEqual(ctx['over'], 'test')

    def test_context_manager_global_behaviour(self):
        ctx = base.Context()
        ctx['global']['x'] = 1

        with ctx as ictx:
            ictx['global']['x'] = 2

        self.assertEqual(ctx['global'], {'x': 2})

    def test_context_manager_local_behaviour(self):
        ctx = base.Context()
        ctx['local']['x'] = 1

        with ctx as ictx:
            ictx['local']['y'] = 1

        self.assertEqual(ctx['local'], {'x': 1})
        self.assertEqual(ictx['local'], {'y': 1})



"""
Context usage:

.. code:: python

    ctx = Context(value=1)
    with ctx as c:
        call_a_function(ctx=c)

A bit more readable version:

.. code:: python

    ctx = Context(value=1)
    with ctx.copywith(value=2, new_value=3) as c:
        call_a_function(ctx=c)
"""
import contextlib


class Context(dict):
    """
    Context provide basic functionality managing through calls.
    Can be used as a context manager.

    Predefined items: global and local

    .. code:: python

        ctx = Context(value=1, global={'value': 1}, local={'value': 1})

        with ctx.copywith(value=2, new_value=3) as c:
            ctx['global']['value']  # contains 1
            ctx['local']  # It's an empty dict
            call_a_function(ctx=c)

    """

    def __init__(self, *args, **kwargs):
        super(Context, self).__init__(*args, **kwargs)
        if 'global' not in self:
            self['global'] = {}
        if 'local' not in self:
            self['local'] = {}

    def __enter__(self):
        cp = self.copy()
        cp['local'] = {}
        return cp

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @contextlib.contextmanager
    def copywith(self, *args, **kwargs):
        """Make a copy of itself and pass back to inside context.

        **Cauction:** It uses simple a `dict.copy` means and if any item itself
        mutable, there is reflecting outside from the context manager as well.
        *Use immutable values if it's possible!*
        """
        with self as copied:
            copied.update(*args, **kwargs)
            yield copied
